package main

import (
	"context"
	"errors"
	"flag"
	"fmt"
	"math"
	"net/http"
	"os"
	"os/signal"
	"sync"
	"time"

	"k8s.io/klog"
)

// daemon is the main state of the succdaemon.
type daemon struct {
	// adcPirani is the adc implementation returning the voltage of the Pfeiffer
	// Pirani gauge.
	adcPirani adc

	// mu guards state variables below.
	mu sync.RWMutex
	// adcPiraniVolts is a moving window of read ADC values, used to calculate a
	// moving average.
	adcPiraniVolts []float32
}

// process runs the pain acquisition and control loop of succd.
func (d *daemon) process(ctx context.Context) {
	ticker := time.NewTicker(time.Millisecond * 100)
	defer ticker.Stop()
	for {
		select {
		case <-ticker.C:
			if err := d.processOnce(ctx); err != nil {
				if errors.Is(err, ctx.Err()) {
					return
				} else {
					klog.Errorf("Processing error: %v", err)
					time.Sleep(time.Second * 10)
				}
			}
		case <-ctx.Done():
			return
		}
	}
}

// processOnce runs the main loop step of succd.
func (d *daemon) processOnce(_ context.Context) error {
	v, err := d.adcPirani.Read()
	if err != nil {
		return fmt.Errorf("when reading ADC: %w", err)
	}
	d.mu.Lock()
	d.adcPiraniVolts = append(d.adcPiraniVolts, v)
	trim := len(d.adcPiraniVolts) - 100
	if trim > 0 {
		d.adcPiraniVolts = d.adcPiraniVolts[trim:]
	}
	d.mu.Unlock()

	return nil
}

// pirani returns the Pirani gauge voltage and pressure.
func (d *daemon) pirani() (volts float32, mbar float32) {
	d.mu.RLock()
	volts = 0.0
	for _, v := range d.adcPiraniVolts {
		volts += v
	}
	if len(d.adcPiraniVolts) != 0 {
		volts /= float32(len(d.adcPiraniVolts))
	}
	d.mu.RUnlock()

	// Per Pirani probe docs.
	bar := math.Pow(10.0, float64(volts)-8.5)
	mbar = float32(bar * 1000.0)
	return
}

var (
	flagFake       bool
	flagListenHTTP string
)

func main() {
	flag.BoolVar(&flagFake, "fake", false, "Enable fake mode which allows to run succd for tests outside the succbone")
	flag.StringVar(&flagListenHTTP, "listen_http", ":8080", "Address at which to listen for HTTP requests")
	flag.Parse()

	ctx, _ := signal.NotifyContext(context.Background(), os.Interrupt)

	d := daemon{}
	if flagFake {
		klog.Infof("Starting with fake Pirani probe")
		d.adcPirani = &fakeADC{}
	} else {
		adc, err := newBBADC(0)
		if err != nil {
			klog.Exitf("Failed to setup Pirani ADC: %v", err)
		}
		d.adcPirani = adc
	}

	http.HandleFunc("/", d.httpIndex)
	http.HandleFunc("/stream", d.httpStream)
	http.HandleFunc("/metrics", d.httpMetrics)

	klog.Infof("Listening for HTTP at %s", flagListenHTTP)
	go func() {
		if err := http.ListenAndServe(flagListenHTTP, nil); err != nil {
			klog.Errorf("HTTP listen failed: %v", err)
		}
	}()

	go d.process(ctx)
	<-ctx.Done()
}
